/*-
 * Copyright 2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#define _POSIX_C_SOURCE 200809L

#include <stdlib.h>
#include <string.h>

#include "config.h"

#include "vcl.h"
#include "vrt.h"
#include "vas.h"
#include "vdef.h"
#include "cache/cache.h"
#include "vsb.h"

#include "vcc_if.h"
#include "vmod_blobdigest.h"

#define ERR(ctx, msg) \
	errmsg((ctx), "vmod blobdigest error: " msg)

#define VERR(ctx, fmt, ...) \
        errmsg((ctx), "vmod blobdigest error: " fmt, __VA_ARGS__)

#define VERRNOMEM(ctx, fmt, ...) \
	VERR((ctx), fmt ", out of space", __VA_ARGS__)

#define ERRNOMEM(ctx, msg) \
	ERR((ctx), msg ", out of space")

#define INIT_FINI(ctx) (((ctx)->method & (VCL_MET_INIT | VCL_MET_FINI)) != 0)

#define ILLEGAL(ctx, m) \
	ERR((ctx), m " is illegal in vcl_init() and vcl_fini().")

struct digest_task {
	unsigned	magic;
#define VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC 0x646937a8
	hash_ctx	ctx;
	struct vmod_priv *result;
};

struct vmod_blobdigest_digest {
	unsigned	magic;
#define VMOD_BLOBDIGEST_DIGEST_MAGIC 0xaccb2e25
	hash_ctx	ctx;
	char		*vcl_name;
	struct vmod_priv *result;
	enum algorithm	hash;
};

struct vmod_blobdigest_hmac {
	unsigned	magic;
#define VMOD_BLOBDIGEST_HMAC_MAGIC 0x85678153
	hash_ctx	inner_ctx;
	hash_ctx	outer_ctx;
	char		*vcl_name;
	enum algorithm	hash;
};

static void
errmsg(VRT_CTX, const char *fmt, ...)
{
	va_list args;

	va_start(args, fmt);
	if (ctx->method == VCL_MET_INIT) {
		AN(ctx->msg);
		VSB_vprintf(ctx->msg, fmt, args);
		VRT_handling(ctx, VCL_RET_FAIL);
	}
	else if (ctx->vsl)
		VSLbv(ctx->vsl, SLT_VCL_Error, fmt, args);
	else
		/* Should this ever happen in vcl_fini() ... */
		VSL(SLT_VCL_Error, 0, fmt, args);
	va_end(args);
}

static void
init(const enum algorithm hash, hash_ctx * const hctx)
{
	switch(hash) {
	case MD5:
		rhash_md5_init(&hctx->md5);
		break;
	case SHA1:
		rhash_sha1_init(&hctx->sha1);
		break;
	case SHA224:
		rhash_sha224_init(&hctx->sha224);
		break;
	case SHA256:
		SHA256_Init(&hctx->sha256);
		break;
	case SHA384:
		rhash_sha384_init(&hctx->sha512);
		break;
	case SHA512:
		rhash_sha512_init(&hctx->sha512);
		break;
	case SHA3_224:
		rhash_sha3_224_init(&hctx->sha3);
		break;
	case SHA3_256:
		rhash_sha3_256_init(&hctx->sha3);
		break;
	case SHA3_384:
		rhash_sha3_384_init(&hctx->sha3);
		break;
	case SHA3_512:
		rhash_sha3_512_init(&hctx->sha3);
		break;
	default:
		WRONG("illegal algorithm");
	}
}

static void
update(const enum algorithm hash, hash_ctx *restrict const hctx,
       const uint8_t *restrict const msg, const size_t len)
{
	switch(hash) {
	case MD5:
		rhash_md5_update(&hctx->md5, msg, len);
		break;
	case SHA1:
		rhash_sha1_update(&hctx->sha1, msg, len);
		break;
	case SHA224:
		rhash_sha256_update(&hctx->sha224, msg, len);
		break;
	case SHA256:
		SHA256_Update(&hctx->sha256, msg, len);
		break;
	case SHA384:
	case SHA512:
		rhash_sha512_update(&hctx->sha512, msg, len);
		break;
	case SHA3_224:
	case SHA3_256:
	case SHA3_384:
	case SHA3_512:
		rhash_sha3_update(&hctx->sha3, msg, len);
		break;
	default:
		WRONG("illegal algorithm");
	}
}

static void
final(const enum algorithm hash, hash_ctx *restrict const hctx,
      uint8_t *restrict result)
{
	switch(hash) {
	case MD5:
		rhash_md5_final(&hctx->md5, result);
		break;
	case SHA1:
		rhash_sha1_final(&hctx->sha1, result);
		break;
	case SHA224:
		rhash_sha256_final(&hctx->sha224, result);
		break;
	case SHA256:
		SHA256_Final(result, &hctx->sha256);
		break;
	case SHA384:
	case SHA512:
		rhash_sha512_final(&hctx->sha512, result);
		break;
	case SHA3_224:
	case SHA3_256:
	case SHA3_384:
	case SHA3_512:
		rhash_sha3_final(&hctx->sha3, result);
		break;
	default:
		WRONG("illegal algorithm");
	}
}

static inline void
digest(const enum algorithm hash, hash_ctx *restrict const hctx,
       VCL_BLOB restrict const b, uint8_t *restrict digest)
{
	init(hash, hctx);
	update(hash, hctx, b->priv, b->len);
	final(hash, hctx, digest);
}

static struct vmod_priv *
ws_alloc_digest(VRT_CTX, const size_t digestsz,
		const char * const restrict context,
		const char * const restrict caller)
{
	struct vmod_priv *b;
	char *snap;

	CHECK_OBJ_NOTNULL(ctx->ws, WS_MAGIC);
	snap = WS_Snapshot(ctx->ws);
	if ((b = WS_Alloc(ctx->ws, sizeof(struct vmod_priv))) == NULL) {
		VERRNOMEM(ctx, "allocating blob in %s.%s()", context, caller);
		return NULL;
	}
	if ((b->priv = WS_Alloc(ctx->ws, digestsz)) == NULL) {
		WS_Reset(ctx->ws, snap);
		VERRNOMEM(ctx, "allocating hash result in %s.%s()", context,
			  caller);
		return NULL;
	}
	b->len = digestsz;
	b->free = NULL;
	return b;
}

/* Objects */

static void
free_digest_task(void *p)
{
        struct digest_task *task;
        CAST_OBJ_NOTNULL(task, p, VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC);
        FREE_OBJ(task);
}

static struct digest_task *
get_task(const struct vrt_ctx * const restrict ctx,
	 const struct vmod_blobdigest_digest * const restrict h)
{
	struct vmod_priv *priv;
	struct digest_task *task;

	priv = VRT_priv_task(ctx, (void *)h);
	AN(priv);
	if (priv->priv != NULL)
		CAST_OBJ(task, priv->priv, VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC);
	else {
		ALLOC_OBJ(task, VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC);
		AN(task);
		memcpy(&task->ctx, &h->ctx, sizeof(hash_ctx));
		AZ(task->result);
		priv->priv = task;
		priv->free = free_digest_task;
	}
	return task;
}

VCL_VOID
vmod_digest__init(VRT_CTX, struct vmod_blobdigest_digest **digestp,
		  const char *vcl_name, VCL_ENUM hashs, VCL_BLOB initb)
{
	struct vmod_blobdigest_digest *digest;
	enum algorithm hash = parse_algorithm(hashs);

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(digestp);
	AZ(*digestp);
	AN(vcl_name);
	ALLOC_OBJ(digest, VMOD_BLOBDIGEST_DIGEST_MAGIC);
	AN(digest);
	*digestp = digest;

	digest->hash = hash;
	digest->vcl_name = strdup(vcl_name);
	AN(digest->vcl_name);
	AZ(digest->result);
	init(hash, &digest->ctx);
	if (initb != NULL)
		update(hash, &digest->ctx, initb->priv, initb->len);
}

VCL_VOID
vmod_digest__fini(struct vmod_blobdigest_digest **digestp)
{
	struct vmod_blobdigest_digest *digest;

	digest = *digestp;
	*digestp = NULL;
	CHECK_OBJ_NOTNULL(digest, VMOD_BLOBDIGEST_DIGEST_MAGIC);
	if (digest->result != NULL) {
		AN(digest->result->priv);
		free(digest->result->priv);
		free(digest->result);
	}
	if (digest->vcl_name != NULL)
		free(digest->vcl_name);
	FREE_OBJ(digest);
}

VCL_BOOL
vmod_digest_update(VRT_CTX, struct vmod_blobdigest_digest *h, VCL_BLOB b)
{
	struct digest_task *task;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(h, VMOD_BLOBDIGEST_DIGEST_MAGIC);

	if (h->result != NULL) {
		VERR(ctx, "already finalized in %s.update()", h->vcl_name);
		return 0;
	}

	if (b == NULL) {
		VERR(ctx, "null BLOB passed to %s.update()", h->vcl_name);
		return 0;
	}

	if (INIT_FINI(ctx)) {
		if (b->len > 0 && b->priv != NULL)
			update(h->hash, &h->ctx, b->priv, b->len);
		return 1;
	}

	task = get_task(ctx, h);
	CHECK_OBJ_NOTNULL(task, VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC);
	if (task->result != NULL) {
		VERR(ctx, "already finalized in %s.update()", h->vcl_name);
		return 0;
	}
	if (b->len > 0 && b->priv != NULL)
		update(h->hash, &task->ctx, b->priv, b->len);
	return 1;
}

VCL_BLOB
vmod_digest_final(VRT_CTX, struct vmod_blobdigest_digest *h)
{
	struct vmod_priv *b;
	struct digest_task *task;
	enum algorithm hash;
	size_t digestsz;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(h, VMOD_BLOBDIGEST_DIGEST_MAGIC);

	if (h->result != NULL)
		return h->result;

	hash = h->hash;
	digestsz = hashspec[hash].digestsz;

	if (INIT_FINI(ctx)) {
		b = malloc(sizeof(struct vmod_priv));
		AN(b);
		b->priv = malloc(digestsz);
		AN(b->priv);
		b->len = digestsz;
		b->free = NULL;
		final(hash, &h->ctx, b->priv);
		h->result = b;
		return b;
	}

	task = get_task(ctx, h);
	CHECK_OBJ_NOTNULL(task, VMOD_BLOBDIGEST_DIGEST_TASK_MAGIC);
	if (task->result != NULL)
		return task->result;

	b = ws_alloc_digest(ctx, digestsz, h->vcl_name, "final");
	if (b == NULL)
		return NULL;
	final(hash, &task->ctx, b->priv);
	task->result = b;
	return b;
}

VCL_VOID
vmod_hmac__init(VRT_CTX, struct vmod_blobdigest_hmac **hmacp,
		const char *vcl_name, VCL_ENUM hashs, VCL_BLOB key)
{
	struct vmod_blobdigest_hmac *hmac;
	enum algorithm hash = parse_algorithm(hashs);
	size_t blocksz = hashspec[hash].blocksz;
	uint8_t k[blocksz], innerk[blocksz], outerk[blocksz];

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	AN(hmacp);
	AZ(*hmacp);
	AN(vcl_name);
	ALLOC_OBJ(hmac, VMOD_BLOBDIGEST_HMAC_MAGIC);
	AN(hmac);
	*hmacp = hmac;
	hmac->hash = hash;
	hmac->vcl_name = strdup(vcl_name);

	memset(k, 0, blocksz);
	if (key->len <= blocksz)
		memcpy(k, key->priv, key->len);
	else {
		hash_ctx hctx[1];

		assert(blocksz >= hashspec[hash].digestsz);
		digest(hash, hctx, key, k);
	}
	for (int i = 0; i < blocksz; i++) {
		innerk[i] = k[i] ^ 0x36;
		outerk[i] = k[i] ^ 0x5c;
	}
	init(hash, &hmac->inner_ctx);
	init(hash, &hmac->outer_ctx);
	update(hash, &hmac->inner_ctx, innerk, blocksz);
	update(hash, &hmac->outer_ctx, outerk, blocksz);
}

VCL_BLOB
vmod_hmac_hmac(VRT_CTX, struct vmod_blobdigest_hmac *h, VCL_BLOB msg)
{
	struct vmod_priv *b;
	hash_ctx inner_ctx, outer_ctx;
	enum algorithm hash;
	size_t digestsz;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	CHECK_OBJ_NOTNULL(h, VMOD_BLOBDIGEST_HMAC_MAGIC);
	hash = h->hash;
	digestsz = hashspec[hash].digestsz;

	b = ws_alloc_digest(ctx, digestsz, h->vcl_name, "hmac");
	if (b == NULL)
		return NULL;

	uint8_t inner_digest[digestsz];
	memcpy(&inner_ctx, &h->inner_ctx, sizeof(hash_ctx));
	memcpy(&outer_ctx, &h->outer_ctx, sizeof(hash_ctx));

	/* Hash the message with the inner key */
	update(hash, &inner_ctx, msg->priv, msg->len);
	final(hash, &inner_ctx, inner_digest);

	/* Hash the result with the outer key */
	update(hash, &outer_ctx, inner_digest, digestsz);
	final(hash, &outer_ctx, b->priv);
	return b;
}

VCL_VOID
vmod_hmac__fini(struct vmod_blobdigest_hmac **hmacp)
{
	struct vmod_blobdigest_hmac *hmac;

	AN(*hmacp);
	hmac = *hmacp;
	*hmacp = NULL;
	CHECK_OBJ_NOTNULL(hmac, VMOD_BLOBDIGEST_HMAC_MAGIC);
	if (hmac->vcl_name != NULL)
		free(hmac->vcl_name);
	FREE_OBJ(hmac);
}

/* Functions */

VCL_BLOB
vmod_hash(VRT_CTX, VCL_ENUM hashs, VCL_BLOB msg)
{
	enum algorithm hash = parse_algorithm(hashs);
	struct vmod_priv *b;
	hash_ctx hctx[1];

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (INIT_FINI(ctx)) {
		ILLEGAL(ctx, "hash()");
		return NULL;
	}
	if (msg == NULL)
		return NULL;

	b = ws_alloc_digest(ctx, hashspec[hash].digestsz, "blobdigest", "hash");
	if (b == NULL)
		return NULL;
	digest(hash, hctx, msg, b->priv);
	return b;
}

VCL_STRING
vmod_version(VRT_CTX __attribute__((unused)))
{
	return VERSION;
}
